import os

# PypI
from PIL import Image

# local
from .classes_consts import ANSI_RGB
from .textAsset import TextAsset


class PictureConverter:
    def __init__(self):
        self._cache_path = os.path.join("__flore1cache__","_pic_cache.jus")

        if not os.path.exists("__flore1cache__"):
            os.makedirs("__flore1cache__", exist_ok=True)

        open(self._cache_path, 'a').close()

        with open(self._cache_path, "r+") as f:
            if os.path.getsize(self._cache_path) == 0:
                f.write(str(dict()))
                self._cache = dict()
            else:
                self._cache = eval(f.read())

    def pics2assets(self, dir="", size="AUTO", alpha=(-1, -1, -1)):
        assets = []
        if not os.path.exists(dir): return assets

        for file in sorted(os.listdir(dir)):
            filename = os.fsdecode(file)
            if filename.endswith(".png") or filename.endswith(".jpg") or filename.endswith(".gif"):
                asset_path = os.path.join(dir, filename)
                assets.append(self.pic2asset(asset_path, size, alpha))

        return assets

# ------------------------------------------------------------

    def pic2asset(self, path="", size="AUTO", alpha=(-1, -1, -1)):
        key = str((path, size, alpha))
        if key in self._cache.keys():
            return TextAsset(self._cache[key])

        img_file = Image.open(path)
        [xs, ys] = img_file.size
        building_manual = []

        scale_filt = Image.NEAREST
        if not path.endswith(".png"):
            scale_filt = Image.LANCZOS

        if size == "AUTO":
            while (xs > 128 * (xs / ys) and ys > 128):
                xs = int(xs / 1.3)
                ys = int(ys / 1.3)
        else:
            xs = int(size[0])
            ys = int(size[1])

        img_file = img_file.resize((int(xs), int(ys)), scale_filt)
        img = img_file.convert("P")
        rgb_palette = self.get_palette_in_rgb(img)

        trsprt_index = self.find_trsprt_index(rgb_palette, alpha)

        if len(rgb_palette) < 255:
            found_black = False
            i = 0
            while i < len(rgb_palette):
                if rgb_palette[i] == (0, 0, 0) and found_black == False:
                    found_black = True
                elif rgb_palette[i] == (0, 0, 0) and found_black == True:
                    del rgb_palette[i]
                    i -= 1
                i += 1

        img = img.load()
        ansi_palette = []
        for color in rgb_palette:
            r, g, b = color
            c_array = [r, g, b]
            nearest = self.nearest_rgb_to_ansi(c_array, ANSI_RGB)
            ansi_palette.append(nearest)

        pv_code = -1
        code = 0
        for y in range(0, ys):
            building_manual.append("")
            for x in range(0, xs):
                ci = img[x, y]
                code = ansi_palette[ci]
                #print("%s%2s" % (("\33["+str(y)+";"+str((x*2)+1)+"H"),"\33[38;5;"+str(ci)+"m@@"))

                if code != pv_code:
                    pixel = "bc:" + str(code) + "  "
                else:
                    pixel = "  "

                if trsprt_index is not None:
                    if rgb_palette[ci] == alpha:
                        pixel = "ªª"

                pv_code = code
                building_manual[y] += pixel

            pv_code = -1

        self._cache[key] = building_manual

        with open(self._cache_path, "w") as f:
            f.truncate(0)
            f.write(str(self._cache))

        return TextAsset(building_manual)

# ------------------------------------------------------------

    def find_trsprt_index(self, rgb_palette, transparent_RGB):
        if transparent_RGB in rgb_palette:
            return rgb_palette.index(transparent_RGB)
        else:
            return None

# ------------------------------------------------------------

    def nearest_rgb_to_ansi(self, c, ref):
        dist_list = []
        c[0] -= 0 # in case you want to filter things up
        c[1] += 0
        c[2] += 0
        for c2 in ref:
            distR = abs(c2[0] - c[0])
            distG = abs(c2[1] - c[1])
            distB = abs(c2[2] - c[2])
            dist_list.append(distR + distG + distB)
        return dist_list.index(min(dist_list))

# ------------------------------------------------------------

    # note: def chunk and def get_palette_in_rgb found on stackoverflow, not my code
    def chunk(self, seq, size, group_by_list=True):
        func = tuple
        if group_by_list:
            func = list
        return [func(seq[i:i + size]) for i in range(0, len(seq), size)]

# ------------------------------------------------------------

    def get_palette_in_rgb(self, img):
        assert img.mode == 'P'
        pal = img.getpalette()
        colors = self.chunk(pal, 3, False)
        return colors

# ------------------------------------------------------------



"""/////////////////////////
///
///   File: flore1.py
///   Author: Anicet Nougaret
///   License: CC BY-SA (see FLORE1/license.txt)
///
/////////////////////////"""

# Python native
import ctypes
import sys
import math
import time
import os
from functools import lru_cache

# pypI
import keyboard
import clipboard

# local
from .engine_classes import *

# -----------------------------------------------------------

def resize_console(col, row):
    os.system('mode con: cols=' + str(col) + ' lines=' + str(row) + '')

def create_dir_and_file(dir, name):
    path = os.path.join(dir, name)
    if not os.path.exists(dir):
        os.makedirs(dir, exist_ok=True)
    open(path, 'a').close()
    return path

def save_var_as(var, dir, filename, ext):
    if not filename.endswith(ext):
        path = create_dir_and_file(dir, filename+ext)
    else:
        path = create_dir_and_file(dir, filename)
    with open(path, "r+") as f:
        if os.path.getsize(path) == 0:
            f.write(str(var))
    return path

def load_var_from(path):
    if not os.path.exists(path):
        print("PATH DOES NOT EXIST")
        time.sleep(2)
        return None
    with open(path, "r") as f:
        if os.path.getsize(path) <= 999999999:
            return eval(f.read())
        else:
            print("FILE TOO BIG")
            time.sleep(2)
            return None

def save_asset_as(asset, folder, name):
    path = os.path.join(folder, name)
    if not hasattr(asset, "chart") or not hasattr(asset, "prtcrd"): return
    save_var_as((asset.chart, asset.prtcrd), folder, name, ".jus")

def load_asset_from(path):
    chart, prtcrd = load_var_from(path)
    if chart != None:
        ta = TextAsset()
        ta.chart = chart
        ta.prtcrd = prtcrd
        return ta
    else: return TextAsset("")

def load_assets_from(folder):
    assets = []
    if not os.path.exists(folder): return []
    for file in sorted(os.listdir(folder)):
        if not file.endswith(".jus"): continue
        path = os.path.join(folder, file)
        chart, prtcrd = load_var_from(path)
        if chart != None:
            ta = TextAsset()
            ta.chart = chart
            ta.prtcrd = prtcrd
            assets.append(ta)
        else: continue
    return assets

def fix_color():
    kernel32 = ctypes.windll.kernel32
    kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

def show_logo(duration=5):
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'engine_assets_lib\logo.png')
    PicConv = PictureConverter()
    scene = Scene(1 , 1, 64, 40, 1)
    logo_asset = PicConv.pic2asset(path=filename, size=[32,32], alpha=(0,0,0))
    logo_sprt = TextSprite(asset=logo_asset)
    scene.put(logo_sprt,32,4,0)
    scene.show()
    print_crd("\033[38;5;8mmade with \033[38;5;198m  F L O R E 1    E N G I N E  ", 43, 3)
    time.sleep(duration)
    print("\33[0m")

def clear_screen():
    print(chr(27) + "[H" + chr(27) + "[J")

def cursor_to_crd(col, row):
    sys.stdout.write("\33[0m\033["+str(row)+";"+str(col)+"H")

def print_crd(content, col, row):
    sys.stdout.write("\33[0m\033["+str(row)+";"+str(col)+"H")
    sys.stdout.write(content)
    sys.stdout.flush()

def save_cursor():
    sys.stdout.write("\033[s")
    sys.stdout.flush()

def restore_cursor():
    sys.stdout.write("\033[u")
    sys.stdout.flush()

def extract_ints(string):
    list = []
    el = ""
    for char in string:
        if char.isdigit():
            el += char
        if not char.isdigit():
            if el != "":
                list.append(int(el))
                el = ""
    return list

def draw_rectangle(char, topleft_x, topleft_y, bottright_x, bottright_y):
    x1 = int(topleft_x)
    y1 = int(topleft_y)
    x2 = int(bottright_x)
    y2 = int(bottright_y)

    for y in range(y1, y1 + (y2 - y1)):
        string = ""
        for x in range(x1, x1 + (x2 - x1)):
            string += char
        print_crd(string, x1, y)

def draw_line(char, xa, ya, xb, yb):
    xa = int(xa); ya = int(ya); xb = int(xb); yb = int(yb)
    if xa == xb and ya == yb: return print_crd(char, xa, ya)

    xa, xb, ya, yb = (xa, xb, ya, yb) if xa < xb else (xb, xa, yb, ya)

    a = (yb - ya) / (xb - xa) if xb != xa else -2

    if -1 <= a <= 1:
        for x in range(0, xb - xa):
            print_crd(char, x+xa, math.floor(x*a + ya))

    if a < -1 or a > 1:
        xa, xb, ya, yb = (xa, xb, ya, yb) if ya > yb else (xb, xa, yb, ya)

        a = (xa - xb) / (ya - yb)

        for y in range(0, ya - yb):
            print_crd(char, math.floor(y*a + xb), y+yb)

def sys_print(string):
    sys.stdout.write(string)
    sys.stdout.flush()

def get_hotkey():
    if not hasattr(get_hotkey, "last"):
        get_hotkey.last = {
            "hotkey": None,
            "delay": 0,
            "time": time.time()
        }
    hotkey = str(keyboard.read_hotkey(suppress=False))

    if hotkey == get_hotkey.last["hotkey"]:
        if get_hotkey.last["time"] + get_hotkey.last["delay"] > time.time():
            return None
        else:
            get_hotkey.last["time"] = time.time()
            get_hotkey.last["delay"] /= 1.5
    else:
        get_hotkey.last["hotkey"] = hotkey
        get_hotkey.last["delay"] = 0.3
        get_hotkey.last["time"] = time.time()

    return hotkey

def get_last_hotkey():
    if not hasattr(get_hotkey, "last"):
        get_hotkey.last = {
            "hotkey": None,
            "delay": 0,
            "time": time.time()
        }
    return get_hotkey.last["hotkey"]


def get_key():
    if not hasattr(get_key, "last"):
        get_key.last = {
            "key": None,
            "delay": 0,
            "time": time.time()
        }
    key = str(keyboard.read_key(suppress=False))

    if key == get_key.last["key"]:
        if get_key.last["time"] + get_key.last["delay"] > time.time():
            return None
        else:
            get_key.last["time"] = time.time()
            get_key.last["delay"] /= 1.5
    else:
        get_key.last["key"] = key
        get_key.last["delay"] = 0.3
        get_key.last["time"] = time.time()

    return key

def get_last_key():
    if not hasattr(get_key, "last"):
        get_key.last = {
            "key": None,
            "delay": 0,
            "time": time.time()
        }
    return get_key.last["key"]

def alt_input(x, y, prompt="", maxline=999, maxchar=99999, digit_only=False, max_width=9999):
    string = ""
    key = ""
    crt_y = y
    crt_x = x
    clip = ""
    last = {
        "key": None,
        "delay": 0,
        "time": time.time()
    }
    cursor_to_crd(x, y)

    @lru_cache(maxsize=32)
    def line_count(string):
        count = 1
        for char in string:
            if char == "\n": count += 1
        return count

    def longest_line(string):
        max = 0
        count = 0
        for char in string:
            count += 1
            if char == "\n":
                if count > max: max += count - max
                count = 0
        return max

    sys.stdout.write(prompt)
    sys.stdout.flush()

    while key != "enter":
        time.sleep(0.05)

        if keyboard.is_pressed("ctrl+c"):
            clipboard.copy(clip)
            continue
        if keyboard.is_pressed("ctrl+v"):
            pasted = clipboard.paste()
            if len(pasted) + len(string) > maxchar: continue
            string += pasted
            crt_x += len(pasted)
            sys.stdout.write(pasted)
            sys.stdout.flush()
            time.sleep(0.2)
            continue

        key = str(keyboard.read_key(suppress=False))

        if key == last["key"]:
            if last["time"] + last["delay"] > time.time():
                continue
            else:
                last["time"] = time.time()
                last["delay"] /= 1.5
        else:
            last["key"] = key
            if last["key"] == "backspace":
                last["delay"] = 0.15
            else:
                last["delay"] = 0.3
            last["time"] = time.time()

        if key == "backspace":
            if crt_x == x: continue
            crt_x -= 1
            if len(string) >= 1: string = string[:len(string)-1]
            sys.stdout.write("\b \b")
            sys.stdout.flush()
        elif key == "suppr":
            while string[len(string)-1] != " " and crt_x > x+1:
                crt_x -= 1
                string = string[:len(string)-1]
                sys.stdout.write("\b \b")
                sys.stdout.flush()
        elif key == "right shift" and line_count(string) < maxline:
            string += "\n"
            crt_y += 1
            crt_x = x
            cursor_to_crd(x, crt_y)
            sys.stdout.flush()
        elif key == "space" and len(string) < maxchar:
            crt_x += 1
            string += " "
            sys.stdout.write(" ")
            sys.stdout.flush()
        elif len(key) == 1 and len(string) < maxchar:
            if digit_only == True and key.isdigit() == False: continue
            crt_x += 1
            string += key
            sys.stdout.write(key)
            sys.stdout.flush()

    draw_rectangle(" ", x, y, x+longest_line(string), y+line_count(string))
    return string



"""/////////////////////////
///
///   File: TEXTASSET/TEXTSPRITE/textSprite.py
///   Author: Anicet Nougaret
///   QuickDesc: The Text Asset's "text sprite" subclass
///   License: CC BY-SA (see FLORE1/license.txt)
///
/////////////////////////"""


# ------------------------------------------------------------
# ------------------   TextSprite Class  ---------------------
# ------------------------------------------------------------
class TextSprite:
  def __init__(self, asset=None):
    self.x = 0
    self.y = 0
    self.layer = None
    self.scene = None
    self.chart = None
    self.prtcrd = set()
    self.act_prtcrd = set()

    if asset != None:
        self.set_asset(asset)

# ------------------------------------------------------------

  def set_asset(self,asset):
    self.chart = asset.chart

    if self.scene != None:
        self.scene.frame_event = True

    if (int(self.x) != 0 or int(self.y) != 0) and asset.prtcrd != self.prtcrd:
      self.prtcrd = asset.prtcrd.copy()
      self.act_prtcrd = set()

      if self.scene != None:
        for px in self.prtcrd:
          x, y = px

          if 0 < x + int(self.x) < self.scene.res_x and 0 < y + int(self.y) < self.scene.res_y:
            self.act_prtcrd.add((x+int(self.x),y+int(self.y)))
    else:
      self.prtcrd = set(asset.prtcrd)
# ------------------------------------------------------------



import math

# ------------------------------------------------------------
# ------------------    FLIPBOOK CLASS   ---------------------
# ------------------------------------------------------------
class Flipbook:
    def __init__(self, Refresh, Sprite, assets, fps=24, sync=True):
        self.Refresh = Refresh
        self.asset_list = assets

        def play(Sprite, asset_list, fps):
            if not hasattr(play, "last_frame"):
                play.last_frame = -1
            if not hasattr(play, "frame"):
                play.frame = 0
            if not hasattr(play, "speed"):
                fps_ratio = play.refresh.fps / fps
                if fps_ratio > 1:
                    play.speed = 1 / fps_ratio
                else:
                    play.speed = fps_ratio

            #print("\033["+str(len(asset_list)-1)+";1H "+("%2s" % str(play.speed)))
            #print("\033["+str(len(asset_list))+";1H "+("%2s" % str(play.last_frame)))

            play.frame = play.speed * play.i
            crt_frame = math.floor(play.frame) % len(asset_list)

            if crt_frame != play.last_frame:
                Sprite.set_asset(asset_list[crt_frame])

            play.last_frame = crt_frame

        play.sync = sync
        self.material = (play, Sprite, self.asset_list, fps)

# ------------------------------------------------------------

    def start(self):
        if not self.Refresh.is_fed_with(*self.material):
            self.Refresh.feed(*self.material)
            return True
        else:
            return False

# ------------------------------------------------------------

    def stop(self):
        if self.Refresh.is_fed_with(*self.material):
            self.Refresh.terminate(*self.material)
            return True
        else:
            return False
# ------------------------------------------------------------



from .flipbook import Flipbook
from .inputHandler import InputHandler
from .refresh import Refresh
from .scene import Scene
from .textAsset import TextAsset
from .textSprite import TextSprite
from .pictureConverter import PictureConverter



"""/////////////////////////
///
///   File: TEXTASSET/textAsset.py
///   Author: Anicet Nougaret
///   QuickDesc: The Engine's subclass managing text assets
///   License: CC BY-SA (see FLORE1/license.txt)
///
/////////////////////////"""

#import time
import os
import sys

# ------------------------------------------------------------
# ------------------    UTIL FUNCTIONS   ---------------------
# ------------------------------------------------------------
def represents_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

# ------------------------------------------------------------

def extract_ansi(word, datalock):
    if word.startswith("bc:"):
        if word[3:] != 256:
            datalock["bc"] = "\u001b[48;5;" + word[3:] + "m \b"
        else:
            datalock["bc"] = None
        if datalock["cc"] == "\33[0m \b":
            datalock["cc"] = None
        return datalock

    if word.startswith("fc:"):
        if word[3:] != 256:
            datalock["fc"] = "\u001b[38;5;" + word[3:] + "m \b"
        else:
            datalock["fc"] = None
        if datalock["cc"] == "\33[0m \b":
            datalock["cc"] = None
        return datalock

    if word.startswith("cc:"):
        if word[3:] == "0":
            datalock["bc"] = None
            datalock["fc"] = None
            datalock["cc"] = "\33[0m \b"
            return datalock

        else:
            datalock["cc"] = "\33[" + word[3:] + "m \b"
            return datalock

# ------------------------------------------------------------

def is_escape_code(word):
    if len(word) > 6: return False

    is_bc = word.startswith("bc:")
    is_fc = word.startswith("fc:")
    is_cc = word.startswith("cc:")
    a = True if is_bc or is_fc or is_cc else False
    if a == False: return False

    b = True

    for char in word[3:]:
        if represents_int(char) == False: b = False

    if b == False: return False

    if (is_bc or is_fc) and not 0 <= int(word[3:]) <= 256: return False
    if (is_cc) and not 0 <= int(word[3:]) <= 1: return False

    return True
# ------------------------------------------------------------


# ------------------------------------------------------------
# ------------------   TextAsset Class   ---------------------
# ------------------------------------------------------------
class TextAsset:
    def __init__(self, building_manual=""):
        self.building_manual = building_manual

        if building_manual != "": self.generate_chart()

# ------------------------------------------------------------

    def generate_chart(self):
        self.chart = {}
        self.prtcrd = set()
        x = 0
        y = 0
        self.datalock = {"bc": None, "fc": None, "cc": None}

        for line in self.building_manual:
            i = -1

            while i < len(line) - 1:
                found_escape = False
                i += 1
                if line[i] != "ª":

                    if i + 6 <= len(line):
                        if is_escape_code(line[i:i + 6]):
                            self.datalock = extract_ansi(line[i:i + 6], self.datalock)
                            i += 5
                            found_escape = True

                    if i + 5 <= len(line) and found_escape == False:
                        if is_escape_code(line[i:i + 5]):
                            self.datalock = extract_ansi(line[i:i + 5], self.datalock)
                            i += 4
                            found_escape = True

                    if i + 4 <= len(line) and found_escape == False:
                        if is_escape_code(line[i:i + 4]):
                            self.datalock = extract_ansi(line[i:i + 4], self.datalock)
                            i += 3
                            found_escape = True

                    if found_escape == False:
                        char = line[i]
                        self.prtcrd.add((x, y))
                        crd = str(x) + "|" + str(y)
                        self.chart[crd] = (char, self.datalock["bc"], self.datalock["fc"], self.datalock["cc"])

                        if self.datalock["bc"] != None:
                            pass
                            #print("\033["+str(y)+";"+str(x)+"H"+self.datalock["bc"]+char)

                        x += 1

                else:
                    x += 1

            x = 0
            y += 1

# ------------------------------------------------------------



import setuptools

setuptools.setup(
    name="flore1",
    version="1.5.0",
    author="Anicet Nougaret",
    author_email="anicet.nougaret@zohomail.eu",
    description="The first 2D python game engine rendering in the terminal.",
    url="https://github.com/AnicetNgrt/flore1",
    packages=setuptools.find_packages(),
    install_requires=[
          'pillow',
	  'keyboard'
      ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Environment :: Console :: Curses",
        "Intended Audience :: Developers"
    ],
    python_requires=' >= 3.7',
)



import time

# ------------------------------------------------------------
# ------------------    REFRESH CLASS    ---------------------
# ------------------------------------------------------------
class Refresh:
    def __init__(self, fps=35):
        self.fps = fps
        self.pv_i = 0
        self.i = 0
        self.pv_frame = 0
        self.frame = 0
        self.stack = []

# ------------------------------------------------------------

    def terminate(self, func, *args, **kwargs):
        self.stack.remove((func, args, kwargs))

# ------------------------------------------------------------

    def is_fed_with(self, func, *args, **kwargs):
        if (func, args, kwargs) in self.stack:
            return True
        else:
            return False

# ------------------------------------------------------------

    def feed(self, func, *args, **kwargs):
        self.stack.append((func, args, kwargs))

# ------------------------------------------------------------

    def do(self):
        for func, args, kwargs in self.stack:
            if hasattr(func, "refresh") == False:
                func.__dict__["refresh"] = self
            if hasattr(func, "sync") == False:
                func.__dict__["sync"] = True
            if not hasattr(func, "i"):
                func.__dict__["i"] = 0
            if not hasattr(func, "last_i"):
                func.__dict__["last_i"] = 0
            else:
                func.__dict__["last_i"] = func.__dict__["i"]

                if func.sync == True and (self.frame-self.pv_frame) > 0:
                    func.__dict__["i"] += self.frame-self.pv_frame
                    func(*args, **kwargs)
                elif func.sync == False:
                    func.__dict__["i"] += 1
                    func(*args, **kwargs)



# ------------------------------------------------------------

    def run(self, debug=False):
        required_fps = self.fps

        start_time = time.time()

        self.do()

        exec_time = time.time()
        latency = exec_time - start_time

        if latency < 0.0001: latency = 0.0001

        fps = 1 / latency

        self.pv_frame = self.frame
        self.pv_i = self.i

        if fps > required_fps:
            time.sleep(((1 / required_fps) - latency)/1.1)
            self.i += 1

        if required_fps >= fps:
            self.i += required_fps / fps

        self.frame = round(self.i)
        #print(self.i)

        if debug:
            otp = "\33[0m\033[1;0H| >>>> Refresh.\33[34mrun \33[33mdebug"
            print(otp)

            otp = "\033[2;0H\33[0m|\u001b[38;5;15m\u001b[48;5;16m  INVERTED_LATENCY_CAP: "
            print(otp + str(self.fps))

            otp = "\33[0m\033[3;0H|\u001b[48;5;16m\u001b[38;5;15m  LATENCY: "
            print("%s%1.2f   " % (otp, latency))

            otp = "\033[4;0H\33[0m|\u001b[38;5;16m"
            if fps >= self.fps:
                otp += "\u001b[48;5;85m"
            elif self.fps * 0.8 <= fps < self.fps:
                otp += "\u001b[48;5;87m"
            elif self.fps * 0.5 <= fps < self.fps * 0.8:
                otp += "\u001b[48;5;221m"
            elif self.fps * 0.2 <= fps < self.fps * 0.5:
                otp += "\u001b[48;5;202m"
            else:
                otp += "\u001b[48;5;9m"
            otp += "  EXECUTION_FREQUENCY: "
            print("%s%1.2f   " % (otp, fps))

            otp = "\33[0m\033[5;0H|\u001b[48;5;16m\u001b[38;5;15m  FRAME_DIFF: "
            print("%s%1.2f   " % (otp, self.frame-self.pv_frame))

            otp = "\33[0m\033[6;0H|\u001b[48;5;16m\u001b[38;5;15m  ACCUMULATOR_DIFF: "
            print("%s%1.2f   " % (otp, self.i-self.pv_i))

            final_time = time.time()
            total_latency = final_time - start_time
            otp = "\33[0m\033[7;0H|\u001b[38;5;15m\u001b[48;5;16m  STABILISED_FREQUENCY: "
            print("%s%1.2f   " % (otp, (self.i - self.pv_i)/total_latency))

# ------------------------------------------------------------



"""/////////////////////////
///
///   File: VIRTUALSCENE/virtualScene.py
///   Author: Anicet Nougaret
///   License: CC BY-SA (see FLORE1/license.txt)
///
/////////////////////////"""

import os
import time
import sys


def print_crd(content, col, row):
    sys.stdout.write("\33[0m\033["+str(row)+";"+str(col)+"H")
    sys.stdout.write(content)
    sys.stdout.flush()

# ------------------------------------------------------------
# ------------------  VirtualScene Class ---------------------
# ------------------------------------------------------------
class Scene:
    def __init__(self, coord_x=0, coord_y=0, res_x=64, res_y=64, layer_count=10, scale=True):
        self.layers = []

        for l in range(0, layer_count):
            self.layers.append([])

        self.coord_x = int(coord_x)
        self.coord_y = int(coord_y)
        self.res_x = int(res_x * 2)
        self.res_y = int(res_y)

        self.prtcrd = set()
        self.prtcrd_rv = set()
        self.pv_prtcrd = set()
        self.pv_prtcrd_rv = set()
        self.chart = set()
        self.pv_chart = set()

        self.frame_event = False

        if scale == True:
            os.system('mode con: cols=' + str(res_x * 2 + coord_x * 4) + ' lines=' + str(res_y + coord_y * 2) + '')

    # ------------------------------------------------------------

    def put(self, sprite, crd_x, crd_y, layer):

        self.frame_event = True

        if layer >= len(self.layers):
            return

        if sprite.scene == self:
            self.erase(sprite)

        sprite.x = crd_x
        sprite.y = crd_y
        sprite.scene = self
        sprite.layer = int(layer)

        sprite.act_prtcrd = set()

        for px in sprite.prtcrd:
            x, y = px
            if 0 < (x + int(sprite.x)) < self.res_x and 0 < (y + int(sprite.y)) < self.res_y:
                sprite.act_prtcrd.add((x + int(sprite.x), y + int(sprite.y)))

        self.layers[layer].append(sprite)

# ------------------------------------------------------------

    def erase(self, sprite):
        self.layers[sprite.layer].remove(sprite)
        self.frame_event = True
        sprite.x = None
        sprite.y = None
        sprite.layer = None
        sprite.scene = None

# ------------------------------------------------------------

    def gen_stream(self, debug, repeat_color):
        self.stream = "\33[0m"
        self.prtcrd = set()
        self.prtcrd_rv = set()
        self.chart = set()

        start_time = time.time()

        sprite_id = 0

        for layer in reversed(self.layers):
            for sprite in layer:
                px_set = sprite.act_prtcrd.copy()

                if not px_set.issubset(self.prtcrd):
                    is_dis = px_set.isdisjoint(self.prtcrd)

                    if not is_dis: px_set.difference_update(self.prtcrd)

                    for px in px_set:
                        x, y = px

                        xp = (y, x)

                        self.prtcrd.add(px)
                        self.prtcrd_rv.add(xp)

                        crd = str(x - int(sprite.x)) + "|" + str(y - int(sprite.y))

                        self.chart.add((xp, sprite.chart[crd]))

        chart_time = time.time()

        self.datalock = {
            "x": None,
            "y": None,
            "bc": None,
            "fc": None,
            "cc": None
        }

        for el in sorted(self.chart - self.pv_chart):
            xp, data = el
            y, x = xp

            char, bc, fc, cc = data

            if x - 1 != self.datalock["x"] or y != self.datalock["y"]:
                self.stream += "\033[" + str(y + self.coord_y) + ";" + str(
                    x + self.coord_x) + "H"
            if repeat_color == False:
                if bc is not None and bc != self.datalock["bc"]:
                    self.stream += bc
                if fc is not None and fc != self.datalock["fc"]:
                    self.stream += fc
                if cc is not None and cc != self.datalock["cc"]:
                    self.stream += cc
            else:
                if bc is not None:
                    self.stream += bc
                if fc is not None:
                    self.stream += fc
                if cc is not None:
                    self.stream += cc

            self.datalock["x"] = x
            self.datalock["y"] = y
            self.datalock["bc"] = bc
            self.datalock["fc"] = fc
            self.datalock["cc"] = cc

            self.stream += char

        stream_time = time.time()

        self.pv_x = None
        self.pv_y = None

        for xp in sorted(self.pv_prtcrd_rv - self.prtcrd_rv):
            y, x = xp

            if True:
                self.stream += "\33[0m" + "\033[" + str(
                    y + self.coord_y) + ";" + str(x + self.coord_x) + "H" + " "
            else:
                self.stream += "\33[0m" + " "

            self.pv_x = x
            self.pv_y = y

        self.pv_prtcrd_rv = self.prtcrd_rv.copy()
        self.pv_chart = self.chart.copy()

        end_time = time.time()

        if debug:
            print("\33[0m\033[1;40H| >>>> Scene \33[33mdebug")
            print("\33[0m\033[2;40H| \33[40m\33[37m FRAME_COMPUTING: %.3f s" % ((chart_time - start_time)))
            print("\33[0m\033[3;40H| \33[40m\33[37m STREAM_COMPUTING: %.3f s" % ((stream_time - chart_time)))
            print("\33[0m\033[4;40H| \33[40m\33[37m ERASE_COMPUTING: %.3f s" % ((end_time - stream_time)))
            print("\33[0m\033[5;40H| \33[40m\33[33m TOTAL_COMPUTING_LATENCY: %.3f s" % ((end_time - start_time)))

# ------------------------------------------------------------

    def show(self, force=False, debug=False, repeat_color=False):
        if self.frame_event == True or force == True:
            genesis_time = time.time()

            if force == True:
                self.pv_prtcrd = set()
                self.pv_prtcrd_rv = set()
                self.pv_chart = set()

            self.gen_stream(debug, repeat_color)

            start_time = time.time()

            sys.stdout.write(self.stream)
            sys.stdout.flush()
            end_time = time.time()
            latency = (end_time - genesis_time)
            if latency == 0: latency = 0.00001

            if debug:
                print("\33[0m\033[6;40H| \33[45m\33[37m SYS.STDOUT.WRITE: %9f s" % ((end_time - start_time)))
                print("\33[0m\033[7;40H| \33[45m\33[37m EXPECTED_FPS: %.2f     " % (round(1 / latency)))

            self.frame_event = False

# ------------------------------------------------------------

    def hide(self):
        self.frame_event = True
        x1 = self.coord_x
        y1 = self.coord_y
        x2 = self.coord_x + self.res_x
        y2 = self.coord_y + self.res_y

        for y in range(y1, y1 + (y2 - y1)):
            string = ""
            for x in range(x1, x1 + (x2 - x1)):
                string += " "
            print_crd(string, x1, y)
# ------------------------------------------------------------

"""
Hey dear stranger,
Thanks for reviewing my code. I hope you will find good use of it.
cheers

Anicet
                                ___
 ,~####~,                    __/__/_
(S O-oO)S  ' , -=o>)',   []-(0_ 0  )--|>>>>>  .  , ~ '~ .\(^ o.o^)

"""
